import operator
from llvmlite import ir
from .omnisci_array import ArrayPointer, Array
from rbc import typesystem
from rbc.utils import get_version
if get_version('numba') >= (0, 49):
    from numba.core import extending, types
else:
    from numba import extending, types


int8_t = ir.IntType(8)
int32_t = ir.IntType(32)
int64_t = ir.IntType(64)
void_t = ir.VoidType()


def overload_binary_cmp_op(op):

    def omnisci_operator_impl(a, e):
        if isinstance(a, ArrayPointer):
            if isinstance(e, ArrayPointer):
                def impl(a, e):
                    if len(a) != len(e):
                        return False
                    for i in range(len(a)):
                        if not op(a[i], e[i]):
                            return False
                    return True
            elif isinstance(e, types.Number):
                def impl(a, e):
                    sz = len(a)
                    x = Array(sz, 'int8')
                    for i in range(sz):
                        x[i] = typesystem.boolean8(op(a[i], e))
                    return x
            return impl

    decorate = extending.overload(op)

    def wrapper(overload_func):
        return decorate(omnisci_operator_impl)

    return wrapper


@overload_binary_cmp_op(operator.eq)
@overload_binary_cmp_op(operator.ne)
@overload_binary_cmp_op(operator.lt)
@overload_binary_cmp_op(operator.le)
@overload_binary_cmp_op(operator.gt)
@overload_binary_cmp_op(operator.ge)
def omnisci_binary_cmp_operator_fn(a, e):
    pass


def overload_binary_op(op, inplace=False):

    def omnisci_operator_impl(a, b):
        if isinstance(a, ArrayPointer) and isinstance(b, ArrayPointer):
            nb_dtype = a.eltype

            def impl(a, b):
                # XXX: raise exception if len(a) != len(b)
                sz = len(a)

                if inplace:
                    x = a
                else:
                    x = Array(sz, nb_dtype)

                for i in range(sz):
                    x[i] = nb_dtype(op(a[i], b[i]))
                return x
            return impl

    decorate = extending.overload(op)

    def wrapper(overload_func):
        return decorate(omnisci_operator_impl)

    return wrapper


@overload_binary_op(operator.add)
@overload_binary_op(operator.and_)
@overload_binary_op(operator.floordiv)
@overload_binary_op(operator.lshift)
@overload_binary_op(operator.mod)
@overload_binary_op(operator.mul)
@overload_binary_op(operator.or_)
@overload_binary_op(operator.pow)
@overload_binary_op(operator.rshift)
@overload_binary_op(operator.sub)
@overload_binary_op(operator.truediv)
@overload_binary_op(operator.xor)
@overload_binary_op(operator.iadd, inplace=True)
@overload_binary_op(operator.iand, inplace=True)
@overload_binary_op(operator.ifloordiv, inplace=True)
@overload_binary_op(operator.ilshift, inplace=True)
@overload_binary_op(operator.imod, inplace=True)
@overload_binary_op(operator.imul, inplace=True)
@overload_binary_op(operator.ior, inplace=True)
@overload_binary_op(operator.ipow, inplace=True)
@overload_binary_op(operator.irshift, inplace=True)
@overload_binary_op(operator.isub, inplace=True)
@overload_binary_op(operator.itruediv, inplace=True)
@overload_binary_op(operator.ixor, inplace=True)
def omnisci_binary_operator_fn(a, b):
    pass


def overload_unary_op(op):

    def omnisci_operator_impl(a):
        if isinstance(a, ArrayPointer):
            nb_dtype = a.eltype

            def impl(a):
                sz = len(a)
                x = Array(sz, nb_dtype)
                for i in range(sz):
                    x[i] = nb_dtype(op(a[i]))
                return x
            return impl

    decorate = extending.overload(op)

    def wrapper(overload_func):
        return decorate(omnisci_operator_impl)

    return wrapper


@overload_unary_op(abs)
@overload_unary_op(operator.abs)
@overload_unary_op(operator.neg)
@overload_unary_op(operator.pos)
def omnisci_unary_op(a):
    pass


@extending.overload(operator.countOf)
def omnisci_array_countOf(a, b):
    """
    Return the number of occurrences of b in a
    """
    if isinstance(a, ArrayPointer) and a.eltype == b:
        def impl_countOf(a, b):
            sz = len(a)
            cnt = 0
            for i in range(sz):
                if a[i] == b:
                    cnt += 1
            return cnt
        return impl_countOf


@extending.lower_builtin(operator.is_, ArrayPointer, ArrayPointer)
def _omnisci_array_is(context, builder, sig, args):
    """Implements `a is b` operation
    """
    [a, b] = args
    return builder.icmp_signed('==', a, b)


@extending.lower_builtin(operator.is_not, ArrayPointer, ArrayPointer)
def _omnisci_array_is_not(context, builder, sig, args):
    """Implements `a is not b` operation
    """
    [a, b] = args
    return builder.icmp_signed('!=', a, b)


@extending.overload(operator.contains)
def omnisci_array_contains(a, e):
    """Implements `e in a` operation
    """
    if isinstance(a, ArrayPointer):
        def impl(a, e):
            sz = len(a)
            for i in range(sz):
                if a[i] == e:
                    return True
            return False
        return impl



import pytest
from numba import njit
from collections import defaultdict
import numpy as np
import math

rbc_omnisci = pytest.importorskip('rbc.omniscidb')
available_version, reason = rbc_omnisci.is_available()
if available_version and available_version < (5, 4):
    reason = ('New-style UDTFs (with Column arguments) are available'
              ' for omniscidb 5.4 or newer, '
              'currently connected to omniscidb '
              + '.'.join(map(str, available_version)))
    available_version = ()
pytestmark = pytest.mark.skipif(not available_version, reason=reason)


def read_csv(filename):
    d = defaultdict(list)
    f = open(filename, 'r')
    s = f.readlines()
    header = s[0].strip().split(',')
    for col in s[1:]:
        col = col.strip().split(',')
        for h, c in zip(header, col):
            d[h].append(float(c))
    return d


@pytest.fixture(scope='module')
def omnisci():
    config = rbc_omnisci.get_client_config(debug=not True)
    m = rbc_omnisci.RemoteOmnisci(**config)
    table_name = 'rbc_test_black_scholes'
    m.sql_execute('DROP TABLE IF EXISTS {table_name}'.format(**locals()))

    sqltypes = ['DOUBLE', 'DOUBLE', 'DOUBLE', 'DOUBLE', 'DOUBLE', 'DOUBLE']
    colnames = ['S', 'X', 'T', 'sigma', 'r', 'oprice']
    table_defn = ',\n'.join('%s %s' % (n, t)
                            for t, n in zip(sqltypes, colnames))
    m.sql_execute(
        'CREATE TABLE IF NOT EXISTS {table_name} ({table_defn});'
        .format(**locals()))

    df = read_csv('./rbc/tests/data_black_scholes.csv')

    rows = 10
    for i in range(rows):
        S = 100.0
        X = df['STRIKE'][i]
        T = df['TIME'][i] / 125
        r = 0.0
        sigma = df['VLTY'][i]
        oprice = df['OPRICE'][i]

        table_row = f'{S}, {X}, {T}, {sigma}, {r}, {oprice}'
        m.sql_execute(
            'INSERT INTO {table_name} VALUES ({table_row})'.format(**locals()))
    m.table_name = table_name
    yield m
    m.sql_execute('DROP TABLE IF EXISTS {table_name}'.format(**locals()))


@njit
def cnd_numba(d):
    A1 = 0.31938153
    A2 = -0.356563782
    A3 = 1.781477937
    A4 = -1.821255978
    A5 = 1.330274429
    RSQRT2PI = 0.39894228040143267793994605993438
    K = 1.0 / (1.0 + 0.2316419 * math.fabs(d))
    ret_val = (RSQRT2PI * math.exp(-0.5 * d * d) *
               (K * (A1 + K * (A2 + K * (A3 + K * (A4 + K * A5))))))
    if d > 0:
        ret_val = 1.0 - ret_val
    return ret_val


def test_black_scholes_udf(omnisci):
    if omnisci.has_cuda and omnisci.version < (5, 5):
        pytest.skip('crashes CUDA enabled omniscidb server'
                    ' [issue 60]')

    omnisci.reset()
    # register an empty set of UDFs in order to avoid unregistering
    # UDFs created directly from LLVM IR strings when executing SQL
    # queries:
    omnisci.register()

    @omnisci('double(double, double, double, double, double)')
    def black_scholes_UDF(S, X, T, r, sigma):
        d1 = (np.log(S/X) + (r + 0.5 * sigma**2)*T) / (sigma * np.sqrt(T))
        d2 = d1 - sigma * np.sqrt(T)

        cndd1 = cnd_numba(d1)
        cndd2 = cnd_numba(d2)

        expRT = math.exp((-1. * r) * T)

        callResult = (S * cndd1 - X * expRT * cndd2)
        return callResult

    S = 100.0
    r = 0.0

    descr, result = omnisci.sql_execute(
        f'select OPRICE, black_scholes_UDF('
        f'cast({S} as DOUBLE), X, T, cast({r} as DOUBLE), sigma)'
        f' from {omnisci.table_name}')

    for _, (expected, out) in enumerate(result):
        assert math.isclose(expected, out, abs_tol=0.0001)


def test_black_scholes_udtf(omnisci):
    if omnisci.has_cuda and omnisci.version < (5, 5):
        pytest.skip('crashes CUDA enabled omniscidb server'
                    ' [issue 169]')

    omnisci.reset()
    # register an empty set of UDFs in order to avoid unregistering
    # UDFs created directly from LLVM IR strings when executing SQL
    # queries:
    omnisci.register()

    @omnisci('int32(Column<double>, Column<double>, Column<double>, Column<double>, Column<double>,'  # noqa: E501
             ' RowMultiplier, OutputColumn<double>)')
    def black_scholes_udtf(S, X, T, r, sigma, m, out):
        input_row_count = len(X)

        for i in range(input_row_count):
            d1 = (np.log(S[i]/X[i]) + (r[i] + 0.5 * sigma[i]**2)*T[i]) / (sigma[i] * np.sqrt(T[i]))  # noqa: E501
            out[i] = S[i] + X[i] + T[i] + r[i] + sigma[i]
            d2 = d1 - sigma[i] * np.sqrt(T[i])

            cndd1 = cnd_numba(d1)
            cndd2 = cnd_numba(d2)

            expRT = math.exp((-1. * r[i]) * T[i])

            out[i] = (S[i] * cndd1 - X[i] * expRT * cndd2)

        return m * input_row_count

    _, result = omnisci.sql_execute(
        'select * from table(black_scholes_udtf('
        ' cursor(select S from {omnisci.table_name}),'
        ' cursor(select X from {omnisci.table_name}),'
        ' cursor(select T from {omnisci.table_name}),'
        ' cursor(select r from {omnisci.table_name}),'
        ' cursor(select sigma from {omnisci.table_name}),'
        ' 1));'
        .format(**locals()))

    _, oprice = omnisci.sql_execute(f'select OPRICE from {omnisci.table_name}')

    for r, o in zip(result, oprice):
        assert math.isclose(o[0], r[0], abs_tol=0.0001)



import pytest
import numpy as np
from rbc.omnisci_backend import Array


rbc_omnisci = pytest.importorskip('rbc.omniscidb')
available_version, reason = rbc_omnisci.is_available()
pytestmark = pytest.mark.skipif(not available_version, reason=reason)


@pytest.fixture(scope='module')
def omnisci():
    config = rbc_omnisci.get_client_config(debug=not True)
    m = rbc_omnisci.RemoteOmnisci(**config)
    # issue https://github.com/xnd-project/rbc/issues/134
    table_name = 'rbc_test_omnisci_array_methods'
    m.sql_execute('DROP TABLE IF EXISTS {table_name}'.format(**locals()))
    yield m


ndarray_methods = [
    ('fill', 'double[](int64, double)', (5, 4), [4.0, 4.0, 4.0, 4.0, 4.0]),
    ('max', 'double(int64, double)', (5, 4.0), 4.0),
    ('max_empty', 'int8(int32)', (0, ), -128),
    ('max_initial', 'double(int64, double, double)', (5, 4.0, 30.0), 30.0),
    ('mean', 'double(int64, double)', (5, 2.0), 2.0),
    ('mean_empty_float', 'float64(int64)', (0, ), np.nan),
    ('mean_empty_int', 'int32(int64)', (0, ), 0),
    ('min', 'double(int64, double)', (5, 4.0), 4.0),
    ('min_empty', 'int16(int64)', (0, ), 32767),
    ('min_initial', 'double(int64, double, double)', (5, 4.0, -3.0), -3.0),
    ('sum', 'double(int64, double)', (5, 2.0), 10.0),
    ('sum_initial', 'double(int64, double, double)', (5, 2.0, 2.0), 12.0),
    ('prod', 'double(int64, double)', (5, 3.0), 243.0),
    ('prod_initial', 'double(int64, double, double)', (5, 3.0, 2), 486.0),
]


def ndarray_fill(size, v):
    a = Array(size, 'double')
    a.fill(v)
    return a


def ndarray_max(size, v):
    a = Array(size, 'double')
    a.fill(v)
    return a.max()


def ndarray_max_empty(size):
    a = Array(size, 'int8')
    return a.max()


def ndarray_max_initial(size, v, initial):
    a = Array(size, 'double')
    a.fill(v)
    return a.max(initial=initial)


def ndarray_mean(size, v):
    a = Array(size, 'double')
    a.fill(v)
    return a.mean()


def ndarray_mean_empty_float(size):
    a = Array(size, 'float64')
    return a.mean()


def ndarray_mean_empty_int(size):
    a = Array(size, 'int32')
    return a.mean()


def ndarray_min(size, v):
    a = Array(size, 'double')
    a.fill(v)
    return a.min()


def ndarray_min_empty(size):
    a = Array(size, 'int16')
    return a.min()


def ndarray_min_initial(size, v, initial):
    a = Array(size, 'double')
    a.fill(v)
    return a.min(initial=initial)


def ndarray_sum(size, v):
    a = Array(size, 'double')
    a.fill(v)
    return a.sum()


def ndarray_sum_initial(size, v, initial):
    a = Array(size, 'double')
    a.fill(v)
    return a.sum(initial=initial)


def ndarray_prod(size, v):
    a = Array(size, 'double')
    a.fill(v)
    return a.prod()


def ndarray_prod_initial(size, v, initial):
    a = Array(size, 'double')
    a.fill(v)
    return a.prod(initial=initial)


@pytest.mark.parametrize("method, signature, args, expected", ndarray_methods,
                         ids=[item[0] for item in ndarray_methods])
def test_ndarray_methods(omnisci, method, signature, args, expected):
    omnisci.reset()

    if omnisci.has_cuda and omnisci.version < (5, 5):
        pytest.skip(
            f'{method}: crashes CUDA enabled omniscidb server [issue 93]')

    if available_version[:3] == (5, 3, 1) and method in ['fill']:
        pytest.skip(
            f'{method}: crashes CPU-only omniscidb server v 5.3.1 [issue 113]')

    if available_version[:3] >= (5, 3, 1) and method in ['max_empty']:
        pytest.skip(
            f'{method}: fails on CPU-only omniscidb server'
            ' v 5.3.1+ [issue 114]')

    omnisci(signature)(eval('ndarray_{}'.format(method)))

    query = 'select ndarray_{method}'.format(**locals()) + \
            '(' + ', '.join(map(str, args)) + ')'
    _, result = omnisci.sql_execute(query)
    out = list(result)[0]

    if method == 'fill':
        assert np.array_equal(expected, out[0]), 'ndarray_' + method
    else:
        assert np.isclose(expected, out, equal_nan=True), \
            'ndarray_' + method



try:
    import numba as nb
    nb_NA_message = ''
except ImportError as msg:
    nb = None
    nb_NA_message = str(msg)

try:
    import numpy as np
    np_NA_message = ''
except ImportError as msg:
    np = None
    np_NA_message = str(msg)

try:
    import numpy as np
except ImportError:
    np = None

import pytest
from rbc.typesystem import Type, get_signature
from rbc.utils import get_datamodel
from rbc.targetinfo import TargetInfo


@pytest.fixture(scope='module')
def target_info():

    target_info = TargetInfo.host()
    with target_info:
        yield target_info


def test_findparen():
    from rbc.typesystem import _findparen as findparen
    assert findparen('a(b)') == 1
    assert findparen('a(b, c())') == 1
    assert findparen('a(b, c(d))') == 1
    assert findparen('a(b(a), c(f(), d))') == 1
    assert findparen('a()(b)') == 3
    assert findparen('{a(),c()}(b)') == 9


def test_commasplit():
    from rbc.typesystem import _commasplit as commasplit
    assert '^'.join(commasplit('a')) == 'a'
    assert '^'.join(commasplit('a, b')) == 'a^b'
    assert '^'.join(commasplit('a, b  , c')) == 'a^b^c'
    assert '^'.join(commasplit('a, (b, e)  , c')) == 'a^(b, e)^c'
    assert '^'.join(commasplit('a, (b, (e,f ,g{h, j}))  , c')) \
        == 'a^(b, (e,f ,g{h, j}))^c'
    assert '^'.join(commasplit('(a, b)')) == '(a, b)'
    assert '^'.join(commasplit('{a, b}')) == '{a, b}'
    assert '^'.join(commasplit('(a, b) , {d, e}')) == '(a, b)^{d, e}'
    assert '^'.join(commasplit('(a[:, :])')) == '(a[:, :])'
    assert '^'.join(commasplit('a[:, :], b[:, :, :]')) == 'a[:, :]^b[:, :, :]'


def test_fromstring(target_info):

    assert Type.fromstring('void') == Type()
    assert Type.fromstring('') == Type()
    assert Type.fromstring('none') == Type()
    assert Type.fromstring('i') == Type('int32')
    assert Type.fromstring('i*') == Type(Type('int32'), '*')
    assert Type.fromstring('*') == Type(Type(), '*')
    assert Type.fromstring('void*') == Type(Type(), '*')
    assert Type.fromstring('{i,j}') == Type(Type('int32'), Type('j'))
    assert Type.fromstring('i(j)') == Type(Type('int32'), (Type('j'), ))
    assert Type.fromstring('i(j , k)') == Type(Type('int32'),
                                               (Type('j'), Type('k')))
    assert Type.fromstring('  (j , k) ') == Type(Type(),
                                                 (Type('j'), Type('k')))
    assert Type.fromstring('void(j,k)') == Type(Type(),
                                                (Type('j'), Type('k')))

    assert Type.fromstring('i a') == Type('int32', name='a')
    assert Type.fromstring('i* a') == Type(Type('int32'), '*', name='a')
    assert Type.fromstring('{i,j} a') == Type(Type('int32'), Type('j'),
                                              name='a')
    assert Type.fromstring('i(j) a') == Type(Type('int32'), (Type('j'),),
                                             name='a')
    assert Type.fromstring('i a*') == Type(Type('int32', name='a'), '*')
    assert Type.fromstring('{i a,j b} c') == Type(Type('int32', name='a'),
                                                  Type('j', name='b'),
                                                  name='c')
    assert Type.fromstring('A<i>') == Type(('A', Type('int32')))
    assert Type.fromstring('A<i*>') == Type(('A', Type(Type('int32'), '*')))
    assert Type.fromstring('A<i>*') == Type(Type(('A', Type('int32'),)), '*')

    with pytest.raises(ValueError, match=r'failed to find lparen index in'):
        Type.fromstring('a)')

    with pytest.raises(ValueError, match=r'failed to comma-split'):
        Type.fromstring('a((b)')

    with pytest.raises(ValueError, match=r'failed to comma-split'):
        Type.fromstring('a((b)')

    with pytest.raises(ValueError, match=r'mismatching curly parenthesis in'):
        Type.fromstring('ab}')


def test_is_properties(target_info):
    t = Type()
    assert t._is_ok and t.is_void
    t = Type('i')
    assert t._is_ok and t.is_atomic
    t = Type('ij')
    assert t._is_ok and t.is_atomic
    t = Type.fromstring('ij')
    assert t._is_ok and t.is_atomic
    t = Type.fromstring('i,j')
    assert t._is_ok and t.is_atomic  # !
    t = Type.fromstring('i  *')
    assert t._is_ok and t.is_pointer
    t = Type.fromstring('*')
    assert t._is_ok and t.is_pointer
    t = Type.fromstring('i* * ')
    assert t._is_ok and t.is_pointer
    t = Type.fromstring('i(j)')
    assert t._is_ok and t.is_function
    t = Type.fromstring('(j)')
    assert t._is_ok and t.is_function
    t = Type.fromstring('()')
    assert t._is_ok and t.is_function
    t = Type.fromstring('{i, j}')
    assert t._is_ok and t.is_struct
    t = Type.fromstring('A<i>')
    assert t._is_ok and t.is_custom

    with pytest.raises(ValueError,
                       match=r'attempt to create an invalid Type object from'):
        Type('a', 'b')


def test_tostring(target_info):

    def tostr(a):
        return Type.fromstring(a).tostring()

    assert tostr('a') == 'a'
    assert tostr('()') == 'void(void)'
    assert tostr('(a,b,c)') == 'void(a, bool, complex64)'
    assert tostr('f  (   )') == 'float32(void)'
    assert tostr('f[a,c]  (   )') == 'f[a,c](void)'
    assert tostr(' f,g ()') == 'f,g(void)'
    assert tostr('a * ') == 'a*'
    assert tostr(' a  * ( b * , c * )  ') == 'a*(bool*, complex64*)'
    assert tostr('{a}') == '{a}'
    assert tostr('{a  ,b}') == '{a, bool}'
    assert tostr('{{a,c} ,b}') == '{{a, complex64}, bool}'
    assert tostr('*') == 'void*'
    assert tostr('void *') == 'void*'
    assert tostr('*(*,{*,*})') == 'void*(void*, {void*, void*})'
    assert tostr('A<a>') == 'A<a>'


def test_normalize(target_info):

    def tostr(a):
        return Type.fromstring(a).tostring()

    assert tostr('a') == 'a'
    assert tostr('int32') == 'int32'
    assert tostr('int32_t') == 'int32'
    assert tostr('int') == 'int32'
    assert tostr('i') == 'int32'
    assert tostr('i35') == 'int35'
    assert tostr('byte') == 'int8'
    assert tostr('ubyte') == 'uint8'

    assert tostr('uint32') == 'uint32'
    assert tostr('uint') == 'uint32'
    assert tostr('u') == 'uint32'
    assert tostr('u35') == 'uint35'
    assert tostr('unsigned int') == 'uint32'

    assert tostr('float32') == 'float32'
    assert tostr('f32') == 'float32'
    assert tostr('f') == 'float32'
    assert tostr('float') == 'float32'
    assert tostr('float64') == 'float64'
    assert tostr('double') == 'float64'
    assert tostr('d') == 'float64'

    assert tostr('complex32') == 'complex32'
    assert tostr('c32') == 'complex32'
    assert tostr('c') == 'complex64'
    assert tostr('complex') == 'complex64'

    assert tostr('') == 'void'
    assert tostr('bool') == 'bool'
    assert tostr('b') == 'bool'
    assert tostr('_Bool') == 'bool'

    assert tostr('str') == 'string'
    assert tostr('string') == 'string'

    assert tostr('i(i*, i15)') == 'int32(int32*, int15)'
    assert tostr('{i,d,c, bool,f,str*}') \
        == '{int32, float64, complex64, bool, float32, string*}'
    assert tostr('A<i(B<f64>)>') == 'A<int32(B<float64>)>'

    datamodel = get_datamodel()
    if datamodel == 'LP64':
        assert tostr('l') == 'int64'
        assert tostr('long') == 'int64'
        assert tostr('long long') == 'int64'
        assert tostr('unsigned long') == 'uint64'
        assert tostr('short') == 'int16'
        assert tostr('unsigned short') == 'uint16'
        assert tostr('ssize_t') == 'int64'
        assert tostr('size_t') == 'uint64'
        assert tostr('c_size_t') == 'uint64'
        assert tostr('std::size_t') == 'uint64'
        assert tostr('long double') == 'float128'
        assert tostr('byte') == 'int8'
        assert tostr('unsigned byte') == 'uint8'
        assert tostr('signed char') == 'int8'
        assert tostr('unsigned char') == 'uint8'
        assert tostr('wchar_t') == 'char32'
        assert tostr('char32') == 'char32'
        assert tostr('signed') == 'int32'
        assert tostr('unsigned') == 'uint32'
    elif datamodel == 'LLP64':
        assert tostr('l') == 'int32'
        assert tostr('long') == 'int32'
        assert tostr('long long') == 'int64'
        assert tostr('unsigned long') == 'uint32'
        assert tostr('short') == 'int16'
        assert tostr('unsigned short') == 'uint16'
        assert tostr('ssize_t') == 'int64'
        assert tostr('size_t') == 'uint64'
        assert tostr('c_size_t') == 'uint64'
        assert tostr('std::size_t') == 'uint64'
        assert tostr('long double') == 'float64'
        assert tostr('byte') == 'int8'
        assert tostr('unsigned byte') == 'uint8'
        assert tostr('signed char') == 'int8'
        assert tostr('unsigned char') == 'uint8'
        assert tostr('wchar_t') == 'char16'
        assert tostr('char32') == 'char32'
        assert tostr('signed') == 'int32'
        assert tostr('unsigned') == 'uint32'
    else:
        raise NotImplementedError('tests for datamodel=`%s`' % (datamodel))


def test_toctypes(target_info):
    import ctypes

    def toctypes(a):
        return Type.fromstring(a).toctypes()

    assert toctypes('bool') == ctypes.c_bool
    assert toctypes('i8') == ctypes.c_int8
    assert toctypes('i32') == ctypes.c_int32
    assert toctypes('u32') == ctypes.c_uint32
    assert toctypes('double') == ctypes.c_double
    assert toctypes('float') == ctypes.c_float
    assert toctypes('char') == ctypes.c_char
    assert toctypes('char8') == ctypes.c_char
    assert toctypes('char*') == ctypes.c_char_p
    assert toctypes('wchar') == ctypes.c_wchar
    assert toctypes('wchar*') == ctypes.c_wchar_p
    assert toctypes('*') == ctypes.c_void_p
    assert toctypes('void*') == ctypes.c_void_p
    assert toctypes('void') is None
    assert toctypes('i(i, double)') \
        == ctypes.CFUNCTYPE(ctypes.c_int32, ctypes.c_int32, ctypes.c_double)
    s = toctypes('{i, double}')
    assert issubclass(s, ctypes.Structure)
    assert s._fields_ == [('f0', ctypes.c_int32), ('f1', ctypes.c_double)]


def test_fromctypes(target_info):
    import ctypes

    def fromstr(a):
        return Type.fromstring(a)

    def fromctypes(t):
        return Type.fromctypes(t)

    assert fromctypes(ctypes.c_char_p) == fromstr('char*')
    assert fromctypes(ctypes.c_wchar_p) == fromstr('wchar*')
    assert fromctypes(ctypes.c_int8) == fromstr('i8')
    assert fromctypes(ctypes.c_uint8) == fromstr('u8')
    assert fromctypes(ctypes.c_uint64) == fromstr('u64')
    assert fromctypes(ctypes.c_float) == fromstr('f32')
    assert fromctypes(ctypes.c_double) == fromstr('double')
    assert fromctypes(ctypes.c_void_p) == fromstr('*')
    assert fromctypes(None) == fromstr('void')

    class mystruct(ctypes.Structure):
        _fields_ = [('f0', ctypes.c_int32), ('f1', ctypes.c_double)]

    assert fromctypes(mystruct) == fromstr('{i32, double}')
    assert fromctypes(ctypes.POINTER(ctypes.c_float)) == fromstr('float*')
    assert fromctypes(ctypes.CFUNCTYPE(ctypes.c_float, ctypes.c_int)) \
        == fromstr('f(i)')


@pytest.mark.skipif(nb is None, reason=nb_NA_message)
def test_tonumba(target_info):
    def tonumba(a):
        return Type.fromstring(a).tonumba()

    assert tonumba('void') == nb.void
    assert tonumba('bool') == nb.boolean
    assert tonumba('int8') == nb.int8
    assert tonumba('int16') == nb.int16
    assert tonumba('int32') == nb.int32
    assert tonumba('int64') == nb.int64
    assert tonumba('uint8') == nb.uint8
    assert tonumba('uint16') == nb.uint16
    assert tonumba('uint32') == nb.uint32
    assert tonumba('uint64') == nb.uint64
    assert tonumba('float') == nb.float32
    assert tonumba('double') == nb.float64
    assert tonumba('complex') == nb.complex64
    assert tonumba('complex128') == nb.complex128
    assert tonumba('double*') == nb.types.CPointer(nb.float64)
    assert tonumba('()') == nb.void()
    assert tonumba('d(i64, f)') == nb.double(nb.int64, nb.float_)
    # assert tonumba('{i,d}')  # numba does not support C struct


@pytest.mark.skipif(nb is None, reason=nb_NA_message)
def test_fromnumba(target_info):
    import numba as nb

    def fromstr(a):
        return Type.fromstring(a)

    def fromnumba(t):
        return Type.fromnumba(t)

    assert fromnumba(nb.void) == fromstr('void')
    assert fromnumba(nb.boolean) == fromstr('bool')
    assert fromnumba(nb.int8) == fromstr('int8')
    assert fromnumba(nb.int16) == fromstr('int16')
    assert fromnumba(nb.int32) == fromstr('int32')
    assert fromnumba(nb.int64) == fromstr('int64')
    assert fromnumba(nb.uint8) == fromstr('uint8')
    assert fromnumba(nb.uint16) == fromstr('uint16')
    assert fromnumba(nb.uint32) == fromstr('uint32')
    assert fromnumba(nb.uint64) == fromstr('uint64')
    assert fromnumba(nb.float_) == fromstr('float32')
    assert fromnumba(nb.double) == fromstr('float64')
    assert fromnumba(nb.complex64) == fromstr('complex64')
    assert fromnumba(nb.complex128) == fromstr('complex128')
    assert fromnumba(nb.types.CPointer(nb.float64)) == fromstr('double*')
    assert fromnumba(nb.double(nb.int64, nb.float_)) == fromstr('d(i64, f)')


@pytest.mark.skipif(np is None, reason='NumPy is not available')
def test_fromnumpy(target_info):

    def fromstr(a):
        return Type.fromstring(a)

    def fromnumpy(t):
        return Type.fromnumpy(t)

    assert fromnumpy(np.void) == fromstr('void')
    assert fromnumpy(np.bool_) == fromstr('bool')
    assert fromnumpy(np.bytes_) == fromstr('bytes')
    assert fromnumpy(np.complex64) == fromstr('complex64')
    assert fromnumpy(np.complex128) == fromstr('complex128')
    assert fromnumpy(np.datetime64) == fromstr('datetime64')
    assert fromnumpy(np.float16) == fromstr('float16')
    assert fromnumpy(np.float32) == fromstr('float32')
    assert fromnumpy(np.float64) == fromstr('float64')
    assert fromnumpy(np.double) == fromstr('float64')
    assert fromnumpy(np.int8) == fromstr('int8')
    assert fromnumpy(np.int16) == fromstr('int16')
    assert fromnumpy(np.int32) == fromstr('int32')
    assert fromnumpy(np.int64) == fromstr('int64')
    assert fromnumpy(np.longlong) == fromstr('int64')
    assert fromnumpy(np.object_) == fromstr('object')
    assert fromnumpy(np.str_) == fromstr('str')
    assert fromnumpy(np.timedelta64) == fromstr('timedelta64')
    assert fromnumpy(np.uint8) == fromstr('uint8')
    assert fromnumpy(np.uint16) == fromstr('uint16')
    assert fromnumpy(np.uint32) == fromstr('uint32')
    assert fromnumpy(np.uint64) == fromstr('uint64')
    assert fromnumpy(np.ulonglong) == fromstr('uint64')

    if hasattr(np, 'float128'):
        assert fromnumpy(np.float128) == fromstr('float128')
    if hasattr(np, 'complex256'):
        assert fromnumpy(np.complex256) == fromstr('complex256')


def test_fromcallable(target_info):

    def foo(a: int, b: float) -> int:
        pass

    assert Type.fromcallable(foo) == Type.fromstring('i64(i64,d)')

    def foo(a: 'int32', b):  # noqa: F821
        pass

    assert Type.fromcallable(foo) == Type.fromstring('void(i32,<type of b>)')

    with pytest.raises(
            ValueError,
            match=(r'constructing Type instance from'
                   r' a lambda function is not supported')):
        Type.fromcallable(lambda a: a)

    with pytest.raises(
            ValueError,
            match=r'callable argument kind must be positional'):
        def foo(*args): pass
        Type.fromcallable(foo)


def test_fromvalue(target_info):
    assert Type.fromvalue(1) == Type.fromstring('i64')
    assert Type.fromvalue(1.0) == Type.fromstring('f64')
    assert Type.fromvalue(1j) == Type.fromstring('c128')
    assert Type.fromvalue("123".encode()) == Type.fromstring('char*')
    assert Type.fromvalue("123") == Type.fromstring('string')
    x = np.dtype(np.float64).type(3.0)
    assert Type.fromvalue(x) == Type.fromstring('float64')
    y = np.dtype(np.complex64).type((1+2j))
    assert Type.fromvalue(y) == Type.fromstring('complex64')


def test_fromobject(target_info):
    import ctypes
    assert Type.fromobject('i8') == Type.fromstring('i8')
    assert Type.fromobject(int) == Type.fromstring('i64')
    assert Type.fromobject(ctypes.c_int16) == Type.fromstring('i16')
    if nb is not None:
        assert Type.fromobject(nb.int16) == Type.fromstring('i16')
    if np is not None:
        assert Type.fromobject(np.int32) == Type.fromstring('i32')
        assert Type.fromobject(np.complex64) == Type.fromstring('complex64')

    def foo():
        pass

    assert Type.fromobject(foo) == Type.fromstring('void(void)')


def test_mangling(target_info):
    def check(s):
        t1 = Type.fromstring(s)
        m = t1.mangle()
        try:
            t2 = Type.demangle(m)
        except Exception:
            print('subject: s=`%s`, t1=`%s`, m=`%s`' % (s, t1, m))
            raise
        assert t1 == t2, repr((t1, m, t2))

    atomic_types = ['void', 'bool', 'char8', 'char16', 'char32',
                    'int8', 'int16', 'int32', 'int64', 'int128',
                    'uint8', 'uint16', 'uint32', 'uint64', 'uint128',
                    'f16', 'f32', 'f64', 'f128',
                    'complex32', 'complex64', 'complex128', 'complex256']
    random_types = ['i8', 'bool', 'f', 'd', '{f}', '{f,d}', '{{f},d}',
                    '{f,{d}}', '{{f},{d}}', '{{{{f}}}}', '()', 'f(d)',
                    'f(())', 'f(d(f))', 'f(d)()', 'f(d)(f(d,d))',
                    '{f}()', '{f,d}({f},f(d,d,d))', 'A<d, d>']
    unknown_types = ['a', 'a*', 'a()', '(a)', 'a(a)', '{a}', '({a})', '{a,a}',
                     'foo', 'bar123', 'V', 'VVV', '_abc_', '_', 'A', 'K', 'P',
                     'A<a>']
    for s in atomic_types + random_types + unknown_types:
        check(s)
        check(s+'*')
        check('{'+s+'}')
        check(s+'('+s+')')
        check('{'+s+','+s+'}')
        check('{'+s+'*,'+s+'}')
        check(s+'('+s+','+s+')')
        check('('+s+','+s+')')
        check(s+'({'+s+'})')
        check(s+'({'+s+'}, '+s+')')
        check(s+'('+s+',{'+s+'})')
        check(s+'('+s+',{'+s+'},'+s+')')
    check('()')


def test_unspecified(target_info):
    assert str(Type.fromstring('unknown(_0,_1)')) == 'unknown(_0, _1)'


def test_annotation(target_info):
    t = Type.fromstring('int foo| a = 1')
    assert t.annotation() == dict(a='1')
    assert t[0] == 'int32'

    def tostr(a):
        return Type.fromstring(a).tostring()

    assert tostr('int foo| a = 1') == 'int32 foo | a=1'
    assert tostr('int foo| a = 1 | b') == 'int32 foo | a=1 | b'
    assert tostr('int foo| a = 1 | a = 2') == 'int32 foo | a=2'

    assert tostr('int| a = 1') == 'int32 | a=1'
    assert tostr('int*| a = 1') == 'int32* | a=1'
    assert tostr('{int, int}| a = 1') == '{int32, int32} | a=1'
    assert (tostr('{int|a=1, int|a=2}| a = 3')
            == '{int32 | a=1, int32 | a=2} | a=3')
    assert tostr('int foo|') == 'int32 foo'
    assert tostr('int foo|a') == 'int32 foo | a'
    assert tostr('int foo|=1') == 'int32 foo | =1'

    t = Type.fromstring('int')
    assert (t | 'a').tostring() == 'int32 | a'
    assert (t | dict(b=1, c=2)).tostring() == 'int32 | b=1 | c=2'


@pytest.mark.skipif(np is None, reason=np_NA_message)
def test_get_signature_ufunc(target_info):

    # Make sure that all get_signature can be applied to all numpy
    # ufuncs
    for name, func in np.__dict__.items():
        if isinstance(func, np.ufunc):
            get_signature(func)

    sig = get_signature(np.trunc)
    assert len(sig.parameters) == 1

    sig = get_signature(np.modf)
    assert len(sig.parameters) == 3



import pytest


rbc_omnisci = pytest.importorskip('rbc.omniscidb')
available_version, reason = rbc_omnisci.is_available()
if available_version and available_version >= (5, 4):
    reason = ('Old-style UDTFs are available for omniscidb 5.3 or older, '
              'currently connected to omniscidb '
              + '.'.join(map(str, available_version)))
    available_version = ()
pytestmark = pytest.mark.skipif(not available_version, reason=reason)


@pytest.fixture(scope='module')
def omnisci():
    config = rbc_omnisci.get_client_config(debug=not True)
    m = rbc_omnisci.RemoteOmnisci(**config)
    table_name = 'rbc_test_omnisci_udtf'
    m.sql_execute('DROP TABLE IF EXISTS {table_name}'.format(**locals()))

    sqltypes = ['FLOAT', 'DOUBLE', 'TINYINT', 'SMALLINT', 'INT', 'BIGINT',
                'BOOLEAN']
    colnames = ['f4', 'f8', 'i1', 'i2', 'i4', 'i8', 'b']
    table_defn = ',\n'.join('%s %s' % (n, t)
                            for t, n in zip(sqltypes, colnames))
    m.sql_execute(
        'CREATE TABLE IF NOT EXISTS {table_name} ({table_defn});'
        .format(**locals()))

    def row_value(row, col, colname):
        if colname == 'b':
            return ("'true'" if row % 2 == 0 else "'false'")
        return row

    rows = 5
    for i in range(rows):
        table_row = ', '.join(str(row_value(i, j, n))
                              for j, n in enumerate(colnames))
        m.sql_execute(
            'INSERT INTO {table_name} VALUES ({table_row})'.format(**locals()))
    m.table_name = table_name
    yield m
    m.sql_execute('DROP TABLE IF EXISTS {table_name}'.format(**locals()))


def test_simple(omnisci):
    omnisci.reset()
    # register an empty set of UDFs in order to avoid unregistering
    # UDFs created directly from LLVM IR strings when executing SQL
    # queries:
    omnisci.register()

    def my_row_copier1(x, input_row_count_ptr, output_row_count, y):
        # sizer type is Constant
        m = 5
        input_row_count = input_row_count_ptr[0]
        n = m * input_row_count
        for i in range(input_row_count):
            for c in range(m):
                y[i + c * input_row_count] = x[i]
        output_row_count[0] = n
        return 0

    if 0:
        omnisci('int32|table(double*|cursor, int64*, int64*, double*|output)')(
            my_row_copier1)
        # Exception: Failed to allocate 5612303629517800 bytes of memory
        descr, result = omnisci.sql_execute(
            'select * from table(my_row_copier1(cursor(select f8 '
            'from {omnisci.table_name})));'
            .format(**locals()))

        for i, r in enumerate(result):
            print(i, r)

    def my_row_copier2(x,
                       n_ptr: dict(sizer='kUserSpecifiedConstantParameter'),  # noqa: F821,E501
                       input_row_count_ptr, output_row_count, y):
        n = n_ptr[0]
        m = 5
        input_row_count = input_row_count_ptr[0]
        for i in range(input_row_count):
            for c in range(m):
                j = i + c * input_row_count
                if j < n:
                    y[j] = x[i]
                else:
                    break
        output_row_count[0] = n
        return 0

    if 0:
        omnisci('int32|table(double*|cursor, int32*|input, int64*, int64*,'
                ' double*|output)')(my_row_copier2)
        # Exception: Failed to allocate 5612303562962920 bytes of memory
        descr, result = omnisci.sql_execute(
            'select * from table(my_row_copier2(cursor(select f8 '
            'from {omnisci.table_name}), 2));'
            .format(**locals()))

        for i, r in enumerate(result):
            print(i, r)

    @omnisci('double(double)')
    def myincr(x):
        return x + 1.0

    @omnisci('int32|table(double*|cursor, int32*|input, int64*, int64*,'
             ' double*|output)')
    def my_row_copier3(x,
                       m_ptr: dict(sizer='kUserSpecifiedRowMultiplier'),  # noqa: F821,E501
                       input_row_count_ptr, output_row_count, y):
        m = m_ptr[0]
        input_row_count = input_row_count_ptr[0]
        for i in range(input_row_count):
            for c in range(m):
                j = i + c * input_row_count
                y[j] = x[i] * 2
        output_row_count[0] = m * input_row_count
        return 0

    descr, result = omnisci.sql_execute(
        'select f8, myincr(f8) from table(my_row_copier3(cursor(select f8 '
        'from {omnisci.table_name}), 2));'
        .format(**locals()))

    for i, r in enumerate(result):
        assert r == ((i % 5) * 2, (i % 5) * 2 + 1)



__all__ = ['omnisci_fixture']


import os
import pytest
from collections import defaultdict


def omnisci_fixture(caller_globals, minimal_version=(0, 0),
                    suffices=['', '10', 'null', 'array', 'arraynull'],
                    load_columnar=True):
    """Usage from a rbc/tests/test_xyz.py file:

      import pytest
      from rbc.tests import omnisci_fixture
      @pytest.fixture(scope='module')
      def omnisci():
          from o in omnisci_fixture(globals()):
              # do some customization here
              yield o

    This fixture creates the following tables:

    f'{omnisci.table_name}' - contains columns f8, f4, i8, i4, i2, i1,
                              b with row size 5.

    f'{omnisci.table_name}10' - contains columns f8, f4, i8, i4, i2,
                                i1, b with row size 10.

    f'{omnisci.table_name}null' - contains columns f8, f4, i8, i4, i2,
                                  i1, b with row size 5, contains null
                                  values.

    f'{omnisci.table_name}array' - contains arrays f8, f4, i8, i4, i2,
                                   i1, b with row size 5

    f'{omnisci.table_name}arraynull' - contains arrays f8, f4, i8, i4, i2,
                                       i1, b with row size 5, contains null
                                       values.
    """
    rbc_omnisci = pytest.importorskip('rbc.omniscidb')
    available_version, reason = rbc_omnisci.is_available()

    def require_version(version, message=None):
        if not available_version:
            pytest.skip(reason)
        if available_version < version:
            _reason = f'test requires version {version} or newer, got {available_version}'
            if message is not None:
                _reason += f': {message}'
            pytest.skip(_reason)

    # Throw an error on Travis CI if the server is not available
    if "TRAVIS" in os.environ and not available_version:
        pytest.exit(msg=reason, returncode=1)

    require_version(minimal_version)

    filename = caller_globals['__file__']
    table_name = os.path.splitext(os.path.basename(filename))[0]

    config = rbc_omnisci.get_client_config(debug=not True)
    m = rbc_omnisci.RemoteOmnisci(**config)

    sqltypes = ['FLOAT', 'DOUBLE', 'TINYINT', 'SMALLINT', 'INT', 'BIGINT',
                'BOOLEAN']
    arrsqltypes = [t + '[]' for t in sqltypes]
    # todo: TEXT ENCODING DICT, TEXT ENCODING NONE, TIMESTAMP, TIME,
    # DATE, DECIMAL/NUMERIC, GEOMETRY: POINT, LINESTRING, POLYGON,
    # MULTIPOLYGON, See
    # https://www.omnisci.com/docs/latest/5_datatypes.html
    colnames = ['f4', 'f8', 'i1', 'i2', 'i4', 'i8', 'b']
    table_defn = ',\n'.join('%s %s' % (n, t)
                            for t, n in zip(sqltypes, colnames))
    arrtable_defn = ',\n'.join('%s %s' % (n, t)
                               for t, n in zip(arrsqltypes, colnames))

    for suffix in suffices:
        m.sql_execute(f'DROP TABLE IF EXISTS {table_name}{suffix}')
        if 'array' in suffix:
            m.sql_execute(f'CREATE TABLE IF NOT EXISTS {table_name}{suffix} ({arrtable_defn});')
        else:
            m.sql_execute(f'CREATE TABLE IF NOT EXISTS {table_name}{suffix} ({table_defn});')

    if load_columnar:
        # fast method using load_table_columnar thrift endpoint, use for large tables
        def row_value(row, col, colname, null=False, arr=False):
            if arr:
                if null and (0 == (row + col) % 2):
                    return None
                a = [row_value(row + i, col, colname, null=null, arr=False) for i in range(row)]
                return a
            if null and (0 == (row + col) % 3):
                return None
            if colname == 'b':
                return row % 2 == 0
            return row

        for suffix in suffices:
            columns = defaultdict(list)
            for j, n in enumerate(colnames):
                for i in range(10 if '10' in suffix else 5):
                    v = row_value(i, j, n, null=('null' in suffix), arr=('array' in suffix))
                    columns[n].append(v)
            m.load_table_columnar(f'{table_name}{suffix}', **columns)

    else:
        # slow method using SQL query statements
        def row_value(row, col, colname, null=False, arr=False):
            if arr:
                if null and (0 == (row + col) % 2):
                    return 'NULL'
                a = [row_value(row + i, col, colname, null=null, arr=False) for i in range(row)]
                return '{' + ', '.join(map(str, a)) + '}'
            if null and (0 == (row + col) % 3):
                return 'NULL'
            if colname == 'b':
                return ("'true'" if row % 2 == 0 else "'false'")
            return row

        for i in range(10):
            if i < 5:
                for suffix in suffices:
                    if suffix == '':
                        table_row = ', '.join(str(row_value(i, j, n))
                                              for j, n in enumerate(colnames))
                    elif suffix == 'null':
                        table_row = ', '.join(str(row_value(i, j, n, null=True))
                                              for j, n in enumerate(colnames))
                    elif suffix == 'array':
                        table_row = ', '.join(str(row_value(i, j, n, arr=True))
                                              for j, n in enumerate(colnames))
                    elif suffix == 'arraynull':
                        table_row = ', '.join(str(row_value(i, j, n, null=True, arr=True))
                                              for j, n in enumerate(colnames))
                    else:
                        continue
                    m.sql_execute(f'INSERT INTO {table_name}{suffix} VALUES ({table_row})')
            if i < 10 and '10' in suffices:
                table_row = ', '.join(str(row_value(i, j, n)) for j, n in enumerate(colnames))
                m.sql_execute(f'INSERT INTO {table_name}10 VALUES ({table_row})')

    m.table_name = table_name
    m.require_version = require_version
    yield m
    for suffix in suffices:
        m.sql_execute(f'DROP TABLE IF EXISTS {table_name}{suffix}')



import atexit
import pytest
import sys
from rbc.remotejit import RemoteJIT, Signature, Caller
from rbc.typesystem import Type

win32 = sys.platform == 'win32'


@pytest.fixture(scope="module")
def rjit(request):
    rjit = RemoteJIT(debug=True)
    rjit.start_server(background=True)
    request.addfinalizer(rjit.stop_server)
    atexit.register(rjit.stop_server)
    return rjit


def with_localjit(test_func):
    ljit = RemoteJIT(local=True)
    device = tuple(ljit.targets)[0]
    target_info = ljit.targets[device]

    def test_func_():
        with target_info:
            return test_func(ljit)
    return test_func_


@with_localjit
def test_construction(ljit):

    assert isinstance(ljit, RemoteJIT)

    # Case 1

    @ljit
    def add(a: int, b: int) -> int:
        return a + b

    assert isinstance(add, Caller)
    signatures = add.get_signatures()
    assert len(signatures) == 1
    assert signatures[0] == Type.fromstring('i64(i64,i64)')

    # Case 2

    @ljit('double(double, double)')
    def add(a, b):
        return a + b

    assert isinstance(add, Caller)
    signatures = add.get_signatures()
    assert len(signatures) == 1
    assert signatures[0] == Type.fromstring('f64(f64,f64)')

    # Case 3

    @ljit('double(double, double)')
    @ljit('int(int, int)')
    def add(a, b):
        return a + b

    assert isinstance(add, Caller)
    signatures = add.get_signatures()
    assert len(signatures) == 2
    assert signatures[1] == Type.fromstring('i32(i32,i32)')
    assert signatures[0] == Type.fromstring('f64(f64,f64)')

    # Case 4

    @ljit('double(double, double)',
          'int(int, int)')
    def add(a, b):
        return a + b

    assert isinstance(add, Caller)
    signatures = add.get_signatures()
    assert len(signatures) == 2
    assert signatures[0] == Type.fromstring('f64(f64,f64)')
    assert signatures[1] == Type.fromstring('i32(i32,i32)')

    # Case 5

    ljit_int = ljit('int(int, int)')
    ljit_double = ljit('double(double, double)')
    assert isinstance(ljit_int, Signature)

    ljit_types = ljit_int(ljit_double)
    assert isinstance(ljit_types, Signature)

    @ljit_types
    def add(a, b):
        return a + b

    assert isinstance(add, Caller)
    signatures = add.get_signatures()
    assert len(signatures) == 2
    assert signatures[0] == Type.fromstring('i32(i32,i32)')
    assert signatures[1] == Type.fromstring('f64(f64,f64)')

    # Case 6

    @ljit
    def add(a, b):
        return a + b

    assert isinstance(add, Caller)
    signatures = add.get_signatures()
    assert len(signatures) == 0

    add.signature('i32(i32,i32)')
    signatures = add.get_signatures()
    assert len(signatures) == 1
    assert signatures[0] == Type.fromstring('i32(i32,i32)')

    # invalid cases are now handled at compile stage


def test_return_scalar(rjit):

    @rjit('i64(i64)', 'f64(f64)', 'c128(c128)')
    def ret(a):
        return a

    r = ret(123)
    assert r == 123
    assert isinstance(r, int)

    r = ret(123.45)
    assert r == 123.45
    assert isinstance(r, float)

    if not win32:
        # see https://github.com/xnd-project/rbc/issues/4
        r = ret(123+45j)
        assert r == 123+45j
        assert isinstance(r, complex)


def test_rjit_add(rjit):

    @rjit('i64(i64,i64)')
    def add(a, b):
        return a + b

    assert isinstance(add(1, 2), int)
    assert add(1, 2) == 3

    with pytest.raises(
            TypeError,
            match=r'found no matching function type to given argument types'):
        add(1, 2.5)

    add.signature('d(d,d)')

    assert isinstance(add(1, 2.5), float)
    assert add(1, 2.5) == 3.5

    assert isinstance(add(1.0, 2.5), float)
    assert add(1.0, 2.5) == 3.5

    with pytest.raises(
            TypeError,
            match=r'found no matching function type to given argument types'):
        add(1j, 2)

    add.signature('c128(c128,c128)')

    if not win32:
        # see https://github.com/xnd-project/rbc/issues/4
        r = add(1j, 2j)
        assert isinstance(r, complex)
        assert r == 3j
        assert add(1j, 2) == 2+1j


def test_options_local(rjit):

    @rjit
    def foo(x: int) -> int:
        return x + 1

    assert foo(1) == 2   # remote execution
    assert foo.local(1) == 2  # local execution


def test_composition(rjit):
    import numba as nb

    @nb.njit
    def fun(x):
        return x*x

    @rjit(local=True)
    def bar(x: int) -> int:
        y = fun(x)
        return y

    # print(bar.get_IR())


@with_localjit
def test_templates(ljit):

    assert isinstance(ljit, RemoteJIT)

    def check_normalized(sig1, sig2):
        sig1 = set(sig1.normalized().signatures)
        sig2 = set(sig2.normalized().signatures)
        assert sig1 == sig2

    # API

    check_normalized(ljit('T(T)', templates=dict(T=['i8', 'i4'])),
                     ljit('i8(i8)', 'i4(i4)'))

    check_normalized(ljit('T(T*)', templates=dict(T=['i8', 'i4'])),
                     ljit('i8(i8*)', 'i4(i4*)'))

    check_normalized(ljit('T({T})', templates=dict(T=['i8', 'i4'])),
                     ljit('i8({i8})', 'i4({i4})'))

    check_normalized(ljit('T(T2)', templates=dict(T=['i8', 'i4'], T2=['i8', 'i4'])),
                     ljit('i8(i8)', 'i4(i4)', 'i8(i4)', 'i4(i8)'))

    check_normalized(ljit('T(T2*)', templates=dict(T=['i8', 'i4'], T2=['i8', 'i4'])),
                     ljit('i8(i8*)', 'i4(i4*)', 'i8(i4*)', 'i4(i8*)'))

    check_normalized(ljit('T({T2})', templates=dict(T=['i8', 'i4'], T2=['i8', 'i4'])),
                     ljit('i8({i8})', 'i4({i4})', 'i8({i4})', 'i4({i8})'))

    check_normalized(ljit('T({T2, T})', templates=dict(T=['i8', 'i4'], T2=['i8', 'i4'])),
                     ljit('i8({i8, i8})', 'i4({i4, i4})', 'i8({i4, i8})', 'i4({i8, i4})'))

    check_normalized(ljit('T(U)', templates=dict(T=['i8', 'i4'], U=['f32'])),
                     ljit('i8(f32)', 'i4(f32)'))

    check_normalized(ljit('T(U)', templates=dict(T=['i8', 'i4'], U=['T'])),
                     ljit('i8(i8)', 'i4(i4)'))

    check_normalized(ljit('T(U)', templates=dict(T=['i8', 'i4'], U=['T*'])),
                     ljit('i8(i8*)', 'i4(i4*)'))

    check_normalized(ljit('T(U)', templates=dict(T=['i8', 'i4'], U=['T*', '{T*}'])),
                     ljit('i8(i8*)', 'i4(i4*)', 'i8({i8*})', 'i4({i4*})'))

    # recursion

    with pytest.raises(
            TypeError,
            match=r"cannot make T concrete"):
        ljit('T(U)', templates=dict(T=['U'], U=['T'])).normalized()

    with pytest.raises(
            TypeError,
            match=r"cannot make T concrete"):
        ljit('T(U)', templates=dict(T=['U'], U=['V'], V=['T'])).normalized()

    # not a concrete type

    with pytest.raises(
            TypeError,
            match=r"cannot make T concrete"):
        ljit('T(T)').normalized()

    # custom types

    class MyClass(Type):
        pass

    class MyClass2(Type):
        pass

    check_normalized(ljit('T(MyClass<T>)', templates=dict(T=['i8', 'i4'])),
                     ljit('i8(MyClass<int8>)', 'int4(MyClass<i4>)'))

    check_normalized(ljit('T(MyClass3<T>)', templates=dict(T=['i8', 'i4'])),
                     ljit('i8(MyClass3<int8>)', 'int4(MyClass3<i4>)'))

    check_normalized(ljit('T(U)', templates=dict(T=['i8', 'i4'], U=['MyClass<T>'])),
                     ljit('i8(MyClass<int8>)', 'int4(MyClass<i4>)'))

    check_normalized(ljit('T(U)', templates=dict(T=['i8', 'i4'], U=['MyClass2<T>'])),
                     ljit('i8(MyClass2<int8>)', 'int4(MyClass2<i4>)'))

    check_normalized(ljit('T(U)', templates=dict(T=['i8'], U=['C<T>'],
                                                 C=['MyClass2', 'MyClass3'])),
                     ljit('i8(MyClass2<int8>)', 'i8(MyClass3<int8>)'))

    check_normalized(ljit('T(U)', templates=dict(T=['i8'], U=['C<T>'],
                                                 C=['MyClass', 'MyClass2', 'MyClass3'])),
                     ljit('i8(MyClass2<int8>)', 'i8(MyClass<int8>)', 'i8(MyClass3<int8>)'))

    # user-friendly interface

    check_normalized(ljit('T(T)', T=['i8', 'i4']),
                     ljit('i8(i8)', 'i4(i4)'))

    check_normalized(ljit('T(U)', T=['i8', 'i4'], U=['MyClass<T>']),
                     ljit('i8(MyClass<int8>)', 'int4(MyClass<i4>)'))

    check_normalized(ljit('T(U)', T=['i8', 'i4'], U=['MyClass3<T>']),
                     ljit('i8(MyClass3<int8>)', 'int4(MyClass3<i4>)'))



from .numpy_funcs import *  # noqa: F401, F403
from .npy_mathimpl import *  # noqa: F401, F403
from .omnisci_array import *  # noqa: F401, F403
from .omnisci_column import *  # noqa: F401, F403
from .omnisci_bytes import *  # noqa: F401, F403
from .numpy_ufuncs import *  # noqa: F401, F403
from .omnisci_pipeline import *  # noqa: F401, F403
from .python_operators import *  # noqa: F401, F403



import pytest
import numpy as np
import rbc.omnisci_backend as omni  # noqa: F401


rbc_omnisci = pytest.importorskip('rbc.omniscidb')
available_version, reason = rbc_omnisci.is_available()
pytestmark = pytest.mark.skipif(not available_version, reason=reason)


@pytest.fixture(scope='module')
def omnisci():
    config = rbc_omnisci.get_client_config(debug=not True)
    m = rbc_omnisci.RemoteOmnisci(**config)
    table_name = 'rbc_test_omnisci_array'
    m.sql_execute('DROP TABLE IF EXISTS {table_name}'.format(**locals()))
    sqltypes = ['FLOAT[]', 'DOUBLE[]',
                'TINYINT[]', 'SMALLINT[]', 'INT[]', 'BIGINT[]',
                'BOOLEAN[]', 'BIGINT[]']
    # todo: TEXT ENCODING DICT, TEXT ENCODING NONE, TIMESTAMP, TIME,
    # DATE, DECIMAL/NUMERIC, GEOMETRY: POINT, LINESTRING, POLYGON,
    # MULTIPOLYGON, See
    # https://www.omnisci.com/docs/latest/5_datatypes.html
    colnames = ['f4', 'f8', 'i1', 'i2', 'i4', 'i8', 'b', 'u8']
    table_defn = ',\n'.join('%s %s' % (n, t)
                            for t, n in zip(sqltypes, colnames))
    m.sql_execute(
        'CREATE TABLE IF NOT EXISTS {table_name} ({table_defn});'
        .format(**locals()))

    def row_value(row, col, colname):
        if colname == 'b':
            return 'ARRAY[%s]' % (', '.join(
                ("'true'" if i % 2 == 0 else "'false'")
                for i in range(-3, 3)))
        if colname.startswith('f'):
            return 'ARRAY[%s]' % (', '.join(
                str(row * 10 + i + 0.5) for i in range(-3, 3)))
        if colname.startswith('u'):
            return 'ARRAY[%s]' % (', '.join(
                str(row * 10 + i) for i in range(6)))
        return 'ARRAY[%s]' % (', '.join(
            str(row * 10 + i) for i in range(-3, 3)))

    rows = 5
    for i in range(rows):
        table_row = ', '.join(str(row_value(i, j, n))
                              for j, n in enumerate(colnames))
        m.sql_execute(
            'INSERT INTO {table_name} VALUES ({table_row})'.format(**locals()))
    m.table_name = table_name
    yield m
    try:
        m.sql_execute('DROP TABLE IF EXISTS {table_name}'.format(**locals()))
    except Exception as msg:
        print('%s in deardown' % (type(msg)))


def num(s):
    try:
        return int(s)
    except ValueError:
        return float(s)


def is_number(s):
    try:
        float(s)
    except ValueError:
        return False
    else:
        return True


binary_fns = [
    ('add', 'int32[](int32[], int32[])', 'i4,i4'),
    ('subtract', 'double[](double[], double[])', 'f8,f8'),
    ('multiply', 'double[](double[], double[])', 'f8,f8'),
    ('divide', 'double[](double[], double[])', 'f8,f8'),
    ('logaddexp', 'double[](double[], double[])', 'f8,f8'),
    ('logaddexp2', 'double[](double[], double[])', 'f8,f8'),
    ('ldexp', 'double[](double[], int64[])', 'f8,i8'),
    ('true_divide', 'double[](double[], double[])', 'f8,f8'),
    ('floor_divide', 'double[](double[], double[])', 'f8,f8'),
    ('power', 'float[](float[], float[])', 'f4,f4'),
    ('remainder', 'double[](double[], double[])', 'f8,f8'),
    ('mod', 'double[](double[], double[])', 'f8,f8'),
    ('fmod', 'double[](double[], double[])', 'f8,f8'),
    ('gcd', 'int64[](int64[], int64[])', 'i8,i8'),
    ('lcm', 'int64[](int64[], int64[])', 'i8,i8'),
    ('arctan2', 'double[](double[], double[])', 'f8,f8'),
    ('hypot', 'double[](double[], double[])', 'f8,f8'),
    ('maximum', 'double[](double[], double[])', 'f8,f8'),
    ('minimum', 'double[](double[], double[])', 'f8,f8'),
    ('greater', 'bool[](int64[], int64[])', 'i8,i8'),
    ('greater_equal', 'bool[](int64[], int64[])', 'i8,i8'),
    ('less', 'bool[](int64[], int64[])', 'i8,i8'),
    ('less_equal', 'bool[](int64[], int64[])', 'i8,i8'),
    ('not_equal', 'bool[](int64[], int64[])', 'i8,i8'),
    ('equal', 'bool[](int64[], int64[])', 'i8,i8'),
    ('logical_and', 'bool[](int64[], int64[])', 'i8,i8'),
    ('logical_or', 'bool[](int64[], int64[])', 'i8,i8'),
    ('logical_xor', 'bool[](int64[], int64[])', 'i8,i8'),
    ('bitwise_and', 'int64[](int64[], int64[])', 'i8,i8'),
    ('bitwise_or', 'int64[](int64[], int64[])', 'i8,i8'),
    ('bitwise_xor', 'int64[](int64[], int64[])', 'i8,i8'),
    ('left_shift', 'int64[](int64[], int64[])', 'u8,u8'),
    ('right_shift', 'int64[](int64[], int64[])', 'i8,i8'),
]


@pytest.mark.parametrize("method, signature, columns", binary_fns,
                         ids=[item[0] for item in binary_fns])
def test_omnisci_array_binary_math(omnisci, method, signature, columns):
    omnisci.reset()

    s = f'def np_{method}(a, b): return omni.{method}(a, b)'
    exec(s, globals())

    omnisci(signature)(eval('np_{}'.format(method)))

    ca, cb = columns.split(',')

    query = f'select {ca}, {cb}, ' + \
            f'np_{method}({ca}, {cb})' + \
            f' from {omnisci.table_name};'

    _, result = omnisci.sql_execute(query)

    row_a, row_b, out = list(result)[0]
    expected = getattr(np, method)(row_a, row_b)
    assert np.isclose(expected, out, equal_nan=True).all(), 'omni_' + method  # noqa: E501


binary_fn_scalar_input = [
    ('add', 'int32[](int32[], int32)', 'i4,3'),
    ('subtract', 'double[](double[], double)', 'f8,5.0'),
    # omnisci server crashes:
    # ('subtract', 'double[](double, double[])', '5.0,f8'),
    ('multiply', 'double[](double[], double)', 'f8,5.0'),
    ('divide', 'double[](double[], double)', 'f8,2.0'),
    ('logaddexp', 'double[](double[], double)', 'f8,2.0'),
    ('logaddexp2', 'double[](double[], double)', 'f8,2.0'),
    ('ldexp', 'double[](double[], int64)', 'f8,3'),
    ('true_divide', 'double[](double[], double)', 'f8,2.0'),
    ('floor_divide', 'double[](double[], double)', 'f8,2.0'),
    ('power', 'int64[](int64[], int64)', 'i8,2'),
    ('remainder', 'double[](double[], double)', 'f8,2.0'),
    ('mod', 'double[](double[], double)', 'f8,2.0'),
    ('fmod', 'double[](double[], double)', 'f8,2.0'),
    ('gcd', 'int64[](int64[], int64)', 'i8,3'),
    ('lcm', 'int64[](int64[], int64)', 'i8,3'),
    ('arctan2', 'double[](double[], double)', 'f8,2.0'),
    ('hypot', 'double[](double[], double)', 'f8,2.0'),
    ('maximum', 'double[](double[], double)', 'f8,2.0'),
    ('minimum', 'double[](double[], double)', 'f8,4.0'),
    ('greater', 'bool[](int64[], int64)', 'i8,1'),
    ('greater_equal', 'bool[](int64[], int64)', 'i8,1'),
    ('less', 'bool[](int64[], int64)', 'i8,1'),
    ('less_equal', 'bool[](int64[], int64)', 'i8,1'),
    ('not_equal', 'bool[](int64[], int64)', 'i8,1'),
    ('equal', 'bool[](int64[], int64)', 'i8,1'),
    ('logical_and', 'bool[](int64[], int64)', 'i8,0'),
    ('logical_or', 'bool[](int64[], int64)', 'i8,2'),
    ('logical_xor', 'bool[](int64[], int64)', 'i8,2'),
    ('bitwise_and', 'int64[](int64[], int64)', 'i8,1'),
    ('bitwise_or', 'int64[](int64[], int64)', 'i8,2'),
    ('bitwise_xor', 'int64[](int64[], int64)', 'i8,2'),
    ('left_shift', 'int64[](int64[], int64)', 'i8,2'),
    ('right_shift', 'int64[](int64[], int64)', 'i8,2'),
]


@pytest.mark.parametrize("method, signature, args", binary_fn_scalar_input,
                         ids=[item[0] for item in binary_fn_scalar_input])
def test_omnisci_array_binary_math_scalar(omnisci, method, signature, args):
    omnisci.reset()

    s = f'def np_{method}(a, b): return omni.{method}(a, b)'
    exec(s, globals())

    omnisci(signature)(eval('np_{}'.format(method)))

    t = tuple(args.split(','))
    a, b = t[0], t[1]
    column = a if is_number(b) else b
    scalar = a if is_number(a) else b

    query = f'select {column}, ' + \
            f'np_{method}({a}, {b})' + \
            f' from {omnisci.table_name};'

    _, result = omnisci.sql_execute(query)

    row, out = list(result)[0]
    expected = getattr(np, method)(row, num(scalar))
    assert np.isclose(expected, out, equal_nan=True).all(), 'omni_' + method  # noqa: E501


unary_fns = [
    # math operations
    ('negative', 'int64[](int64[])', 'i8'),
    # ('positive', 'int64[](int64[])', 'i8'),
    # ('absolute', 'int64[](double[])', 'i8'),
    ('absolute', 'double[](double[])', 'f8'),
    ('fabs', 'double[](double[])', 'f8'),
    ('rint', 'double[](double[])', 'f8'),
    # ('absolute'), 'double[](double[])', 'f8'),
    # ('conj'), 'double[](double[])', 'f8'),
    # ('conjugate'), 'double[](double[])', 'f8'),
    ('exp', 'double[](double[])', 'f8'),
    ('exp2', 'double[](double[])', 'f8'),
    ('log', 'double[](double[])', 'f8'),
    ('log2', 'double[](double[])', 'f8'),
    ('log10', 'double[](double[])', 'f8'),
    ('expm1', 'double[](double[])', 'f8'),
    ('log1p', 'double[](double[])', 'f8'),
    ('sqrt', 'double[](double[])', 'f8'),
    ('square', 'double[](double[])', 'f8'),
    # ('cbrt', 'double[](double[])', 'f8'),
    ('reciprocal', 'double[](double[])', 'f8'),
    # Bit-twiddling functions
    ('invert', 'int64[](int64[])', 'i8'),  # invert does a bitwise inversion
    ('invert', 'bool[](bool[])', 'b'),
    # trigonometric functions
    ('sin', 'double[](double[])', 'f8'),
    ('cos', 'double[](double[])', 'f8'),
    ('tan', 'double[](double[])', 'f8'),
    ('arcsin', 'double[](double[])', 'f8'),
    ('arccos', 'double[](double[])', 'f8'),
    ('arctan', 'double[](double[])', 'f8'),
    ('sinh', 'double[](double[])', 'f8'),
    ('cosh', 'double[](double[])', 'f8'),
    ('tanh', 'double[](double[])', 'f8'),
    ('arcsinh', 'double[](double[])', 'f8'),
    ('arccosh', 'double[](double[])', 'f8'),
    ('arctanh', 'double[](double[])', 'f8'),
    ('deg2rad', 'double[](double[])', 'f8'),
    ('rad2deg', 'double[](double[])', 'f8'),
    # comparison functions
    ('logical_not', 'bool[](int64[])', 'i8'),
    # floating functions
    ('isfinite', 'bool[](int64[])', 'i8'),
    ('isinf', 'bool[](int64[])', 'i8'),
    ('isnan', 'bool[](int64[])', 'i8'),
    # ('isnat', 'bool[](int64[])', 'i8'), # doesn't work
    ('fabs', 'double[](double[])', 'f8'),
    ('signbit', 'bool[](int64[])', 'i8'),
    ('signbit', 'bool[](double[])', 'f8'),
    # ('spacing', 'double[](double[])', 'f8'), # not supported
    ('floor', 'double[](double[])', 'f8'),
    ('ceil', 'double[](double[])', 'f8'),
    ('trunc', 'double[](double[])', 'f8'),
]


@pytest.mark.parametrize("method, signature, column", unary_fns,
                         ids=[item[0] for item in unary_fns])
def test_omnisci_array_unary_math_fns(omnisci, method, signature, column):
    omnisci.reset()

    s = f'def np_{method}(a): return omni.{method}(a)'
    exec(s, globals())

    omnisci(signature)(eval('np_{}'.format(method)))

    query = f'select {column}, ' + \
            f'np_{method}({column})' + \
            f' from {omnisci.table_name};'

    _, result = omnisci.sql_execute(query)

    row, out = list(result)[0]
    if method == 'invert' and column == 'b':
        # invert on booleans is different from invert on int8 values
        row = np.array(row) != 0

    expected = getattr(np, method)(row)
    assert np.isclose(expected, out, equal_nan=True).all(), 'omni_' + method  # noqa: E501


def test_heaviside(omnisci):

    @omnisci('double[](int64[], int64)')
    def heaviside(x1, x2):
        return omni.heaviside(x1, x2)

    query = f'select i8, heaviside(i8, 1) from {omnisci.table_name}'
    _, result = omnisci.sql_execute(query)
    result = list(result)

    for inp, out in result:
        expected = np.heaviside(inp, 1)
        np.array_equal(expected, out)




import os
import pytest
import numpy as np
from rbc.thrift import Server, Client
from rbc.utils import get_local_ip
from .test_thrift_multiplexed import DispatcherTest

test_thrift_file = os.path.join(os.path.dirname(__file__), 'test.thrift')
socket_options = dict(host=get_local_ip(), port=6325+1,
                      thrift_content_service='test', multiplexed=False)


@pytest.fixture(scope="module")
def server(request):
    print('staring rpc_thrift server ...', end='')
    test_thrift_file = os.path.join(os.path.dirname(__file__), 'test.thrift')
    options = dict(multiplexed=False)
    options.update(socket_options)
    ps = Server.run_bg(DispatcherTest, test_thrift_file, options)

    def fin():
        if ps.is_alive():
            print('... stopping rpc_thrift server')
            ps.terminate()

    request.addfinalizer(fin)


def test_buffer_transport(server):
    conn = Client(**socket_options)
    arr = np.array([1, 2, 3, 4], dtype=np.int64)
    r = conn(test=dict(test_buffer_transport=(arr,)))
    np.testing.assert_equal(r['test']['test_buffer_transport'],
                            arr.view(np.uint8))


def test_ndarray_transport(server):
    conn = Client(**socket_options)
    arr = np.array([1, 2, 3, 4], dtype=np.int64)
    r = conn(test=dict(test_ndarray_transport=(arr,)))
    np.testing.assert_equal(r['test']['test_ndarray_transport'], arr)

    arr = np.array([[1, 2, 3, 4], [5, 6, 7, 8]], dtype=np.float64)
    r = conn(test=dict(test_ndarray_transport=(arr,)))
    np.testing.assert_equal(r['test']['test_ndarray_transport'], arr)


def test_ndarray_transport2(server):
    conn = Client(**socket_options)
    arr = np.array([1, 2, 3, 4], dtype=np.int64)
    r = conn(test=dict(test_ndarray_transport2=(arr,)))
    np.testing.assert_equal(r['test']['test_ndarray_transport2'], arr)


def test_Data_transport(server):
    conn = Client(**socket_options)
    data = dict(a=1, b=[1, 2, 3], c=dict(d=13.4))
    r = conn(test=dict(test_Data_transport=(data,)))
    assert r['test']['test_Data_transport'] == data


def test_Data_transport2(server):
    conn = Client(**socket_options)
    data = dict(a=1, b=[1, 2, 3], c=dict(d=13.4))
    r = conn(test=dict(test_Data_transport2=(data,)))
    assert r['test']['test_Data_transport2'] == data


def test_str_transport(server):
    conn = Client(**socket_options)
    s = 'hello'
    r = conn(test=dict(test_str_transport=(s,)))
    assert r['test']['test_str_transport'] == s


def test_bool_transport(server):
    conn = Client(**socket_options)
    s = True
    r = conn(test=dict(test_bool_transport=(s,)))
    assert r['test']['test_bool_transport'] == s

    s = False
    r = conn(test=dict(test_bool_transport=(s,)))
    assert r['test']['test_bool_transport'] == s


def test_byte_transport(server):
    conn = Client(**socket_options)
    s = 8
    r = conn(test=dict(test_byte_transport=(s,)))
    assert r['test']['test_byte_transport'] == s


def test_int16_transport(server):
    conn = Client(**socket_options)
    s = 16
    r = conn(test=dict(test_int16_transport=(s,)))
    assert r['test']['test_int16_transport'] == s


def test_int32_transport(server):
    conn = Client(**socket_options)
    s = 32
    r = conn(test=dict(test_int32_transport=(s,)))
    assert r['test']['test_int32_transport'] == s


def test_int64_transport(server):
    conn = Client(**socket_options)
    s = 64
    r = conn(test=dict(test_int64_transport=(s,)))
    assert r['test']['test_int64_transport'] == s


def test_int64_transport2(server):
    conn = Client(**socket_options)
    s = 64
    r = conn(test=dict(test_int64_transport2=(s,)))
    assert r['test']['test_int64_transport2'] == s


def test_double_transport(server):
    conn = Client(**socket_options)
    s = 3.14
    r = conn(test=dict(test_double_transport=(s,)))
    assert r['test']['test_double_transport'] == s


def test_set_transport(server):
    conn = Client(**socket_options)
    s = set(['abc', 'cde'])
    r = conn(test=dict(test_set_transport=(s,)))
    assert r['test']['test_set_transport'] == s


def test_list_transport(server):
    conn = Client(**socket_options)
    s = [1, 2, 3, 4]
    r = conn(test=dict(test_list_transport=(s,)))
    assert r['test']['test_list_transport'] == s


def test_map_transport(server):
    conn = Client(**socket_options)
    s = {1: 'one', 2: 'two'}
    r = conn(test=dict(test_map_transport=(s,)))
    assert r['test']['test_map_transport'] == s


def test_void(server):
    conn = Client(**socket_options)
    r = conn(test=dict(test_void=()))
    assert r['test']['test_void'] is None


def test_exception(server):
    conn = Client(**socket_options)
    with pytest.raises(
            ValueError,
            match="my exception"):
        conn(test=dict(test_exception=()))


def test_myenum_transport(server):
    conn = Client(**socket_options)
    s = conn.thrift.MyEnum.MyONE
    r = conn(test=dict(test_myenum_transport=(s,)))
    assert r['test']['test_myenum_transport'] == s
    s = conn.thrift.MyEnum.MyTWO
    r = conn(test=dict(test_myenum_transport=(s,)))
    assert r['test']['test_myenum_transport'] == s
